# -*- coding: utf-8 -*-

import pytest

parametrize = pytest.mark.parametrize

# Random text retrieved from lorem ipsum generator. For further reference,
# check http://www.lipsum.com/.
_TEXT = '''Lorem ipsum dolor sit amet, vestibulum dui dictumst mauris in
turpis. Neque augue leo, pellentesque purus neque ornare atque, nec vitae,
tortor justo risus justo. Aptent pellentesque ipsum lacinia lacinia consequat,
at consectetuer eleifend rhoncus, hendrerit et amet erat, quisque mus sem hac
donec et.

Leo ut nullam, condimentum in accumsan eu dolor vestibulum, ac egestas
imperdiet felis amet. Id odio.

'''


@pytest.fixture
def create_temp_folder(tmpdir):
    """Fixture for creating temporary folder for tests."""
    def _create_temp_folder(subdirs, filenames):
        parent_dir = tmpdir.mkdir('parent')

        for subdir in subdirs:
            childdir = parent_dir.mkdir(subdir)

            for filename in filenames:
                fp = childdir.join(filename)
                fp.write(_TEXT)

        return str(parent_dir)
    return _create_temp_folder


@pytest.fixture
def create_temp_file(tmpdir):
    """Fixture for creating a temporary file."""
    def _create_temp_file(filename):
        fp = tmpdir.mkdir('parent').mkdir('child').join(filename)
        fp.write(_TEXT)
        return str(fp)
    return _create_temp_file



# -*- coding: utf-8 -*-

import os
import queue
import threading
import logging as log

from .schema import FinderSchema, DataSchema, ErrorSchema
from .utils import (
    is_kernel_file,
    is_executable,
    is_readable,
    is_image,
    is_audio,
    is_video
)

log.basicConfig(format='%(message)s', level=log.INFO)

lock = threading.Lock


class FileReader(threading.Thread):
    """Thread class for finding the given pattern in a given file.

    Args:
        path (str): File path.
        pattern (str): A string to be searched in the given file path.

    .. versionadded:: 1.0.0
    """
    def __init__(self, path=None, pattern=None, _queue=None):
        self.path = path
        self.pattern = pattern
        self._queue = _queue
        threading.Thread.__init__(self)

    def run(self):
        """Runs methods for reading the file and searching the pattern in a
        line of text from the file.
        """
        items = []
        error = []
        pattern_found = False
        errors_occurred = False

        try:
            for line_number, line in read(self.path):
                if search(line, self.pattern):
                    items.append(self.set_data(line_number, line))
                    pattern_found = True
        except PermissionError as exc:
            error.append(self.make_error(type='PermissionError',
                                         message=exc,
                                         extra=None))
            errors_occurred = True
        except OSError as exc:
            error.append(self.make_error(type='OSError',
                                         message=exc,
                                         extra='File might be a kernel file'))
            errors_occurred = True
        except UnicodeDecodeError as exc:
            # TODO: Support all the file encodings.
            error.append(self.make_error(type='UnicodeDecodeError',
                                         message=exc,
                                         extra=None))
            errors_occurred = True

        if pattern_found or errors_occurred:
            # Serialize to a JSON-encoded string.
            response = FinderSchema().dumps({'path': self.path,
                                             'total_items': len(items),
                                             'items': items,
                                             'error': error})

            self._queue.put(response.data)

    def set_data(self, line_number, line):
        """Returns the serialized data."""
        data = {'line_number': line_number,
                'line': line}

        return DataSchema().dump(data).data

    def make_error(self, type, message, extra):
        """Returns the error serialized data."""
        error = {'type': type,
                 'message': message,
                 'extra': extra}

        return ErrorSchema().dump(error).data


def find(*paths, **kwargs):
    """Main method for finding the pattern in the given file paths.

    Args:
        paths (list): List of paths to be walked through.
        kwargs (str): Positional arguments.

    .. versionadded:: 1.0.0
    """
    pattern = kwargs.get('pattern', None)
    _queue = queue.Queue()

    for path in iterfiles(*paths):
        reader = FileReader(path, pattern, _queue)
        reader.start()
        reader.join()

        while not _queue.empty():
            yield _queue.get()


def search(text=None, pattern=None):
    """Searches the given pattern in the given text.

    Args:
        text (str): Text in which the pattern needs to be searched.
        pattern (str): A string to be searched in the given `text`.

    Returns:
        bool: A boolean value stating whether the pattern is found in the
            given text or not.

    .. versionadded:: 1.0.0
    """
    return True if pattern in text else False


def read(path):
    """Yields line by line from the file path provided.

    Args:
        path (str): File path provided.

    Yields:
        int: Line number.
        str: Line of text corresponding to the line number from the file.

    .. versionadded:: 1.0.0
    """
    with open(path, 'r', encoding='utf-8') as fp:
        count = 0

        for line in fp:
            count += 1
            yield (count, line.strip('\n'))


def iterfiles(*paths):
    """Yields all the non-executable file paths in a given directory.

    Args:
        paths (list): List of paths to be walked through.

    Yields:
        path (str): File path (files in the given directory path).

    .. versionadded:: 1.0.0
    """
    for path in paths:
        path = os.path.expanduser(path) if not os.path.isabs(path) else path

        if not os.path.exists(path):
            log.info("{path} is not a valid path. Please provide a valid path."
                     .format(path=path))
            continue

        if os.path.isfile(path) and not is_executable(path):
            yield path
        elif os.path.isdir(path):
            for root, dirs, files in os.walk(path):
                for file in files:
                    file_path = os.path.join(root, file)

                    if (os.path.isfile(file_path) and
                            is_readable(file_path) and
                            not is_image(file_path) and
                            not is_executable(file_path) and
                            not is_kernel_file(file_path) and
                            not is_audio(file_path) and
                            not is_video(file_path)):
                        yield file_path



# -*- coding: utf-8 -*-

from .fixtures import parametrize, create_temp_file
from finder.api import read


@parametrize('filename', [
    'childfile1.txt'
])
def test_read_file(filename, create_temp_file):
    """Test that validates whether the line count from the file is correct or
    not.
    """
    file_path = create_temp_file(filename)
    total_line_count = max([i for i, j in read(file_path)])

    with open(file_path, 'r') as fp:
        assert len(fp.readlines()) == total_line_count



# -*- coding: utf-8 -*-

import os

from .extensions import (
    IMAGE_FORMATS,
    VIDEO_FORMATS,
    AUDIO_FORMATS,
    KERNEL_DIRS
)


def split_params(params):
    """Returns a list of values from a given string of comma-separated values.
    """
    return params.split(',')


def file_extension(path):
    """Returns the file extension of the given file path."""
    return path.split(os.path.sep)[-1].split('.')[-1]


def is_executable(path):
    """Validates whether a given file path is binary or not.

    Args:
        path (str): File path.

    Returns:
        bool: True if the file is a binary else False.

    .. versionadded:: 1.0.0
    """
    return os.access(path, os.X_OK)


def is_readable(path):
    """Validates whether a given file path is readable or not.

    Args:
        path (str): File path.

    Returns:
        bool: True if the file is readable else False.

    .. versionadded:: 1.0.0
    """
    return os.access(path, os.R_OK)


def is_image(path):
    """Validates whether a given file path is image or not.

    Args:
        path (str): File path.

    Returns:
        bool: True if the file is an image else False.

    .. versionadded:: 1.0.0
    """
    extension = file_extension(path)

    return extension in IMAGE_FORMATS


def is_video(path):
    """Validates whether a given file path is a video or not.

    Args:
        path (str): File path.

    Returns:
        bool: True if the file is a video else False.

    .. versionadded:: 1.0.0
    """
    extension = file_extension(path)

    return extension in VIDEO_FORMATS


def is_audio(path):
    """Validates whether a given file path is a audio or not.

    Args:
        path (str): File path.

    Returns:
        bool: True if the file is a audio else False.

    .. versionadded:: 1.0.0
    """
    extension = file_extension(path)

    return extension in AUDIO_FORMATS


def is_kernel_file(path):
    """Validates whether the file path is a kernel file or not.

    Args:
        path (str): File path.

    Returns:
        bool: True if the file path is a kernel file else False.

    .. versionadded:: 1.0.0
    """
    return any([path.startswith(parent_dir) for parent_dir in KERNEL_DIRS])



# -*- coding: utf-8 -*-

from finder.api import iterfiles
from .fixtures import parametrize, create_temp_folder, create_temp_file


def validate_filenames(filepaths, expected_filenames):
    """Validates the results against the expected filenames."""
    results = list(iterfiles(filepaths))
    assert len(results) == len(expected_filenames)


@parametrize('subdirs,filenames,expected_filenames', [
    (['child1', 'child2'],
     ['childfile1.txt', 'childfile2.txt'],
     ['child1/childfile1.txt',
      'child1/childfile2.txt',
      'child2/childfile1.txt',
      'child2/childfile2.txt'])
])
def test_iterfiles_folder(create_temp_folder,
                          subdirs,
                          filenames,
                          expected_filenames):
    """Test for validating whether the method `iterfiles` can iterate through
    all the file names in a given folder or not.
    """
    folder_path = create_temp_folder(subdirs, filenames)
    validate_filenames(folder_path, expected_filenames)


@parametrize('filename,expected_filenames', [
    ('childfile1.txt', ['childfile1.txt'])
])
def test_iterfiles(create_temp_file, filename, expected_filenames):
    """Test for validating whether the method `iterfiles` can yield a given
    filepath (not a folder path).
    """
    file_path = create_temp_file(filename)
    validate_filenames(file_path, expected_filenames)



# -*- coding: utf-8 -*-

import json

import pytest

from finder.api import find
from .fixtures import create_temp_folder, parametrize


@parametrize('data', [
    {'subdirs': ['child1', 'child2'],
     'filenames': ['childfile1.txt', 'childfile2.txt'],
     'pattern': 'Lorem',
     'expected_total_items': '1',
     'expected_items': [{
         'line_number': '1',
         'line': 'Lorem ipsum dolor sit amet, vestibulum dui dictumst mauris '
                 'in'}]},
    {'subdirs': ['child1', 'child2'],
     'filenames': ['childfile1.txt', 'childfile2.txt'],
     'pattern': 'in',
     'expected_total_items': '3',
     'expected_items': [
         {'line_number': '1',
          'line': 'Lorem ipsum dolor sit amet, vestibulum dui dictumst mauris '
                  'in'},
         {'line_number': '3',
          'line': 'tortor justo risus justo. Aptent pellentesque ipsum '
                  'lacinia lacinia consequat,'},
         {'line_number': '7',
          'line': 'Leo ut nullam, condimentum in accumsan eu dolor vestibulum,'
                  ' ac egestas'}]},
])
def test_find(create_temp_folder, data):
    """Tests that validates whether `find()` is able to retrieve the expected
    line and line number.
    """
    folder_path = create_temp_folder(data['subdirs'], data['filenames'])

    for result in find(folder_path, pattern=data['pattern']):
        _result = json.loads(result)
        assert _result['total_items'] == data['expected_total_items']

        assert _result['items'] == data['expected_items']



# -*- coding: utf-8 -*-
"""Define project metadata."""


__package_name__ = 'finder'
__description__ = ('Command line interface for searching a given pattern in '
                   'the given directory/file paths')
__url__ = 'https://github.com/bharadwajyarlagadda/finder'
__version__ = '1.0.1'
__author__ = 'Bharadwaj Yarlagadda'
__email__ = 'yarlagaddabharadwaj@gmail.com'
__license__ = 'MIT License'



# -*- coding: utf-8 -*-

from datetime import datetime

from marshmallow import Schema, fields

from .__pkg__ import __version__


class DataSchema(Schema):
    """Schema for the data that is searched in the given file paths."""
    line_number = fields.String(default=None)
    line = fields.String(default=None)

    class Meta:
        fields = ('line_number', 'line')
        ordered = True


class ErrorSchema(Schema):
    """Schema for errors."""
    type = fields.String(default=None)
    message = fields.String(default=None)
    extra = fields.String(default=None)

    class Meta:
        fields = ('type', 'message', 'extra')
        ordered = True


class FinderSchema(Schema):
    """Finder API schema."""
    api_version = fields.String(default=__version__)
    requested_on = fields.DateTime(default=datetime.utcnow().isoformat())
    path = fields.String(default=None)
    total_items = fields.String(default=None)
    items = fields.Nested(DataSchema, many=True)
    error = fields.Nested(ErrorSchema, many=True)

    class Meta:
        fields = ('api_version',
                  'requested_on',
                  'path',
                  'total_items',
                  'items',
                  'error')
        ordered = True



# -*- coding: utf-8 -*-

import pytest
from click.testing import CliRunner

from finder.cli import finder
from finder.__pkg__ import __version__

from .fixtures import create_temp_folder, parametrize


@pytest.yield_fixture
def cli():
    runner = CliRunner()
    yield runner


def test_finder_version(cli):
    """Test that validates the version of finder tool."""
    result = cli.invoke(finder, ['--version'])

    assert result.exit_code == 0
    assert result.output.strip() == __version__


@parametrize('data', [
    {'subdirs': ['child1', 'child2'],
     'filenames': ['childfile1.txt', 'childfile2.txt'],
     'args': ['search',
              '--pattern={0}'.format('Lorem')],
     'expected_output':
         [['1', ' Lorem ipsum dolor sit amet, vestibulum dui dictumst mauris '
                'in']]},
    {'subdirs': ['child1', 'child2'],
     'filenames': ['childfile1.txt', 'childfile2.txt'],
     'args': ['search',
              '--pattern={0}'.format('in')],
     'expected_output':
         [['1', ' Lorem ipsum dolor sit amet, vestibulum dui dictumst mauris '
                'in'],
          ['3', ' tortor justo risus justo. Aptent pellentesque ipsum lacinia '
                'lacinia consequat,'],
          ['7', ' Leo ut nullam, condimentum in accumsan eu dolor vestibulum, '
                'ac egestas']]},
    {'subdirs': ['child1', 'child2'],
     'filenames': ['childfile1.txt', 'childfile2.txt'],
     'args': ['search',
              '--pattern={0}'.format('Lorem'),
              '--verbose'],
     'expected_output':
         [['1', ' Lorem ipsum dolor sit amet, vestibulum dui dictumst mauris '
                'in']]},
    {'subdirs': ['child1', 'child2'],
     'filenames': ['childfile1.txt', 'childfile2.txt'],
     'args': ['search',
              '--pattern={0}'.format('in'),
              '--verbose'],
     'expected_output':
         [['1', ' Lorem ipsum dolor sit amet, vestibulum dui dictumst mauris '
                'in'],
          ['3', ' tortor justo risus justo. Aptent pellentesque ipsum lacinia '
                'lacinia consequat,'],
          ['7', ' Leo ut nullam, condimentum in accumsan eu dolor vestibulum, '
                'ac egestas']]},
])
def test_search_normal_view(cli, create_temp_folder, data):
    """Test for validating finder's search functionality."""
    folder_path = create_temp_folder(data['subdirs'], data['filenames'])

    args = data['args'] + ['--path={0}'.format(folder_path)]

    result = cli.invoke(finder, args)
    out = result.output.strip()

    for output in out.split('\n'):
        assert output.split(':')[1:] in data['expected_output']

    assert result.exit_code == 0



# -*- coding: utf-8 -*-

import os
import json

import click

from .__pkg__ import __version__
from .api import find
from .utils import split_params


# Borrowed from http://click.pocoo.org/5/advanced/.
class AliasedGroup(click.Group):

    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: %s' % ', '.join(sorted(matches)))


@click.group(cls=AliasedGroup, invoke_without_command=True)
@click.option('-v', '--version', is_flag=True, default=False,
              help='Finder tool version.')
@click.pass_context
def finder(ctx, version):
    """Command line interface for searching a given pattern/text in the given
    directory/file path.
    """
    if version:
        click.echo(__version__)


@finder.command('search')
@click.option('--path', default=os.getcwd(),
              help='Searches for the pattern in the directory/file path '
                   'provided. If a directory/file path is not provided, the '
                   'tool searches for the pattern in the current working '
                   'directory.',
              metavar='path1[,path2,...,pathN]')
@click.option('pattern', '-P', '--pattern', '--text', required=True,
              help='Text to be searched.',
              metavar='<pattern>')
@click.option('-v', '--verbose', is_flag=True, default=False,
              help='Some files cannot be opened and searched for the given '
                   'pattern. For example, kernel files which generate content '
                   'on go, files which are not utf-8 encoded, etc. You can '
                   'use this flag if you need a detailed output of which file '
                   'has an error.')
@click.pass_context
def search(ctx, path, pattern, verbose):
    """Searches for the given pattern in the directory/file path provided."""
    for result in find(*split_params(path), pattern=pattern):
        _result = json.loads(result)
        path = _result['path']
        items = _result['items']
        errors = _result['error']

        if items:
            for item in items:
                click.echo('{path}:{line_number}: {line}'
                           .format(path=path,
                                   line_number=item['line_number'],
                                   line=item['line']))

        if errors and verbose:
            for error in errors:
                click.echo('{type}:{path}:{message} {extra}'
                           .format(type=error['type'],
                                   path=path,
                                   message=error['message'],
                                   extra=error['extra'] or ''))



# -*- coding: utf-8 -*-

from finder.api import search
from .fixtures import parametrize


@parametrize('kargs,expected', [
    ({'text': 'hello world 1122', 'pattern': 'wor'}, True),
    ({'text': 'hello world 1122', 'pattern': 'w'}, True),
    ({'text': 'hello world 1122', 'pattern': '1122'}, True),
    ({'text': 'hello world 1122', 'pattern': ' '}, True),
    ({'text': 'hello world 1122', 'pattern': 'row'}, False)
])
def test_search_text(kargs, expected):
    """Test for validating the search function."""
    assert search(text=kargs['text'], pattern=kargs['pattern']) == expected



# -*- coding: utf-8 -*-
# All the extensions specified below are non-readable content. It's better to
# avoid these in our search method. This is the place where you can update the
# non-readable file extensions.

# Images are not readable content. Any image extension that is not readable
# need to be added here in this array.
IMAGE_FORMATS = [
    'bmp',
    'dib',
    'tif',
    'tiff',
    'gif',
    'jpe',
    'jpg',
    'jpeg',
    'jif',
    'jfif',
    'jfi',
    'jp2',
    'jpf',
    'jpm',
    'jpx',
    'j2k',
    'j2c',
    'mj2',
    'fpx',
    'pcd',
    'png',
    'pbm',
    'pgm',
    'ppm',
    'pnm',
    'webp',
    'heif',
    'heic',
    'bpg'
]


# Videos are not readable content. Any video extension that is not readable
# need to be added here in this array.
VIDEO_FORMATS = [
    'webm',
    'mkv',
    'flv',
    'vob',
    'ogv',
    'ogg',
    'drc',
    'gifv',
    'mng',
    'avi',
    'mov',
    'qt',
    'wmv',
    'yuv',
    'rm',
    'rmvb',
    'asf',
    'amv',
    'mp4',
    'm4p',
    'm4v',
    'mpg',
    'mp2',
    'mpeg',
    'mpe',
    'mpv',
    'm2v',
    'm4v',
    'svi',
    '3gp',
    '3g2',
    'mxf',
    'roq',
    'nsv',
    'f4v',
    'f4p',
    'f4a',
    'f4b'
]


# Audios are not readable content. Any audio extension that is not readable
# need to be added here in this array.
AUDIO_FORMATS = [
    'aa',
    'aac',
    'aax',
    'act',
    'aiff',
    'amr',
    'ape',
    'au',
    'awb',
    'dct',
    'dss',
    'dvf',
    'flac',
    'gsm',
    'iklax',
    'ivs',
    'm4a',
    'm4b',
    'mmf',
    'mp3',
    'mpc',
    'msv',
    'ogg',
    'oga',
    'mogg',
    'opus',
    'ra',
    'rm',
    'raw',
    'sln',
    'tta',
    'vox',
    'wav',
    'wma',
    'wv'
]


# Some of the kernel files are dynamically generated based on the process
# running such as the files in '/proc' filesystem. These files are basically
# kernel files and the content of these files is generated on the go.
# Searching these files would be complex and the memory would be running out
# if we read these files. It is better we avoid searching these directories.
KERNEL_DIRS = ['/proc']
